"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.CanvasPickerPlugin = exports.PointInPathPickerFactory = void 0;

var _tslib = require("tslib");

var _g = require("@antv/g");

var _gPluginCanvasRenderer = require("@antv/g-plugin-canvas-renderer");

var _glMatrix = require("gl-matrix");

var _inversify = require("inversify");

var PointInPathPickerFactory = 'PointInPathPicker';
/**
 * pick shape(s) with Mouse/Touch event
 *
 * 1. find AABB with r-tree
 * 2. do math calculation with geometry in an accurate way
 */

exports.PointInPathPickerFactory = PointInPathPickerFactory;

var CanvasPickerPlugin = function () {
  function CanvasPickerPlugin() {
    var _this = this;

    this.isHit = function (displayObject, position, worldTransform) {
      // use picker for current shape's type
      var pick = _this.pointInPathPickerFactory(displayObject.nodeName);

      if (pick) {
        // invert with world matrix
        var invertWorldMat = _glMatrix.mat4.invert(_glMatrix.mat4.create(), worldTransform); // transform client position to local space, do picking in local space


        var localPosition = _glMatrix.vec3.transformMat4(_glMatrix.vec3.create(), _glMatrix.vec3.fromValues(position[0], position[1], 0), invertWorldMat); // account for anchor


        var halfExtents = displayObject.getGeometryBounds().halfExtents;
        var _a = displayObject.parsedStyle.anchor,
            anchor = _a === void 0 ? [0, 0] : _a;
        localPosition[0] += anchor[0] * halfExtents[0] * 2;
        localPosition[1] += anchor[1] * halfExtents[1] * 2;

        if (pick(displayObject, new _g.Point(localPosition[0], localPosition[1]), _this.isPointInPath)) {
          return true;
        }
      }

      return false;
    };
    /**
     * use native picking method
     * @see https://developer.mozilla.org/zh-CN/docs/Web/API/CanvasRenderingContext2D/isPointInPath
     */


    this.isPointInPath = function (displayObject, position) {
      var context = _this.offscreenCanvas.getOrCreateContext();

      var generatePath = _this.pathGeneratorFactory(displayObject.nodeName);

      if (generatePath) {
        generatePath(context, displayObject.parsedStyle);
      }

      return context.isPointInPath(position.x, position.y);
    };
  }

  CanvasPickerPlugin_1 = CanvasPickerPlugin;

  CanvasPickerPlugin.prototype.apply = function (renderingService) {
    var _this = this;

    renderingService.hooks.pick.tap(CanvasPickerPlugin_1.tag, function (result) {
      // position in world space
      var _a = result.position,
          x = _a.x,
          y = _a.y;

      var position = _glMatrix.vec3.fromValues(x, y, 0);

      var pickedDisplayObjects = []; // query by AABB first with spatial index(r-tree)

      var rBushNodes = _this.rBush.search({
        minX: position[0],
        minY: position[1],
        maxX: position[0],
        maxY: position[1]
      });

      var queriedNames = rBushNodes.map(function (node) {
        return node.name;
      });
      rBushNodes.forEach(function (_a) {
        var name = _a.name;

        var displayObject = _this.displayObjectPool.getByName(name);

        if (displayObject.isVisible() && displayObject.interactive) {
          // parent is not included, eg. parent is clipped
          if (displayObject.parentNode && queriedNames.indexOf(displayObject.parentNode.getEntity().getName()) === -1) {
            return;
          } // test with clip path


          var objectToTest = displayObject.style.clipPath || displayObject;
          var worldTransform = displayObject.getWorldTransform(); // clipped's world matrix * clipPath's local matrix

          if (displayObject.style.clipPath) {
            worldTransform = _glMatrix.mat4.multiply(_glMatrix.mat4.create(), worldTransform, displayObject.style.clipPath.getLocalTransform());
          }

          if (_this.isHit(objectToTest, position, worldTransform)) {
            pickedDisplayObjects.push(displayObject);
          }
        }
      }); // find group with max z-index

      pickedDisplayObjects.sort(_this.sceneGraphService.sort);
      return {
        position: result.position,
        // return last picked
        picked: pickedDisplayObjects[pickedDisplayObjects.length - 1] || null
      };
    });
  };

  var CanvasPickerPlugin_1;
  CanvasPickerPlugin.tag = 'CanvasPickerPlugin';
  (0, _tslib.__decorate)([(0, _inversify.inject)(_g.DefaultCamera), (0, _tslib.__metadata)("design:type", _g.Camera)], CanvasPickerPlugin.prototype, "camera", void 0);
  (0, _tslib.__decorate)([(0, _inversify.inject)(_g.SceneGraphService), (0, _tslib.__metadata)("design:type", Object)], CanvasPickerPlugin.prototype, "sceneGraphService", void 0);
  (0, _tslib.__decorate)([(0, _inversify.inject)(_g.RenderingContext), (0, _tslib.__metadata)("design:type", Object)], CanvasPickerPlugin.prototype, "renderingContext", void 0);
  (0, _tslib.__decorate)([(0, _inversify.inject)(_g.DisplayObjectPool), (0, _tslib.__metadata)("design:type", _g.DisplayObjectPool)], CanvasPickerPlugin.prototype, "displayObjectPool", void 0);
  (0, _tslib.__decorate)([(0, _inversify.inject)(_g.OffscreenCanvasCreator), (0, _tslib.__metadata)("design:type", _g.OffscreenCanvasCreator)], CanvasPickerPlugin.prototype, "offscreenCanvas", void 0);
  (0, _tslib.__decorate)([(0, _inversify.inject)(_gPluginCanvasRenderer.RBushRoot), (0, _tslib.__metadata)("design:type", Object)], CanvasPickerPlugin.prototype, "rBush", void 0);
  (0, _tslib.__decorate)([(0, _inversify.inject)(_gPluginCanvasRenderer.PathGeneratorFactory), (0, _tslib.__metadata)("design:type", Function)], CanvasPickerPlugin.prototype, "pathGeneratorFactory", void 0);
  (0, _tslib.__decorate)([(0, _inversify.inject)(PointInPathPickerFactory), (0, _tslib.__metadata)("design:type", Function)], CanvasPickerPlugin.prototype, "pointInPathPickerFactory", void 0);
  CanvasPickerPlugin = CanvasPickerPlugin_1 = (0, _tslib.__decorate)([(0, _inversify.injectable)()], CanvasPickerPlugin);
  return CanvasPickerPlugin;
}();

exports.CanvasPickerPlugin = CanvasPickerPlugin;